var cqrParser = {
    allowedMIMEs: [],
    reSearchBody: new RegExp('<search.+?<\/search>', 'ig'),
                            //matches first <search ...> ... </search>
    reSerchTag: new RegExp('<search\\s.+?>(?=[^"<]+?<|$)', 'ig'),
                            //matches all <search params="">
    reInputTag: new RegExp('<input\\s.+?>(?=[^"<]+?<|$)', 'ig'),
                            //matches all <input params="">
    reTagParams: new RegExp('\\w+\\s*?=\\s*?".*?"', 'ig'),
                            //matches all params=""
    reUserParam: /\s+?user(="")*(?=\s*[\/>])/ig,
                            //matches Mycroft <input param="" user>
    reComments: new RegExp('#.{0,}', 'igm'),
                            //removes #comments in file globally
    reSpecials: new RegExp('[\n\r\f\t]', 'igm'), 
                            //linirizes plugin file

    init: function cqrParser_init(){
        amArr = cqrEnv.AllowedMIMETypesInOrder.split(",");
        for (each in amArr) cqrParser.allowedMIMEs.push(cqrTools.trim(amArr[each]));
        cqrLog.debug('Parsers Factory loaded.');
    },


    ///////////////////////
    // XML parsers section
    //
    parseXML: function parseXML(filedata, itemObject, omit_inputs){
        try{
            //var retItem = new Object();
            var parser = new DOMParser();
            var doc = parser.parseFromString(filedata, "text/xml");
            var roottag = doc.documentElement;
            
            if ((roottag.tagName == "parserError") ||
                (roottag.namespaceURI == "http://www.mozilla.org/newlayout/xml/parsererror.xml")){
                cqrLog.err(itemObject.filepath + ' XML data malformed!', roottag.textContent);
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = roottag.textContent;
                return itemObject;
            }
            if (!((roottag.namespaceURI == "http://www.mozilla.org/2006/browser/search/") ||
                 (roottag.namespaceURI == "http://geek.net.ru/en/opensource/conquery/")  ||
                 (roottag.namespaceURI == "http://a9.com/-/spec/opensearch/1.1/"))) {
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = 'XML is not in ConQuery, MozillaSearch or OpenSearch namespaces!\nGot: ' + roottag.namespaceURI;
                cqrLog.err(itemObject.filepath, itemObject.Diagnosis);
                return itemObject;
            }
            
            if (roottag.tagName.toLowerCase() == 'conqueryplugin'){
                return cqrParser.xmlConQuery(roottag, itemObject, 'ConQuery');
            } else if (roottag.tagName.toLowerCase() == 'searchplugin'){
                return cqrParser.xmlMozilla(roottag, itemObject, 'MozillaSearch');
            } else if (roottag.tagName.toLowerCase() == 'opensearchdescription'){
                return cqrParser.xmlOpenSearch(roottag, itemObject, 'OpenSearch');
            } else {
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = 'Unknown root tag, got: ' + roottag.tagName;
                cqrLog.err(itemObject.filepath, itemObject.Diagnosis);
                return itemObject;
            }
        } catch (err) {
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseXML cant parse plugin!\n' + err;
            cqrLog.err(filepath, itemObject.Diagnosis);
            return itemObject;
        }
    },
    
    xmlConQuery: function xmlConQuery(node, itemObject, plugType){
        itemObject.PluginType = plugType;
        itemObject.Title = cqrParser.xmlGetFirst(node, "ShortName", cqrMess.GetStringFromName('cqr.err.parse'));
        itemObject.Desc = cqrParser.xmlGetFirst(node, "Description");
        itemObject.qcharset = cqrParser.xmlGetFirst(node, "InputEncoding");
        itemObject.raw_icon = cqrParser.xmlGetFirst(node, "Image");
        if (!itemObject.raw_icon) {
            itemObject.iconpath = cqrParser.xmlGetFirst(node, "IconUpdateUrl");
        }
        itemObject.referrer = cqrParser.xmlGetFirst(node, "referrer");
        itemObject.opentarget = cqrParser.xmlGetFirst(node, "opentarget");
        itemObject.openwin = cqrParser.xmlGetFirst(node, "openwin");
        itemObject.forcein = cqrParser.xmlGetFirst(node, "forcein");
        itemObject.showif = cqrParser.xmlGetFirst(node, "showif");
        if (itemObject.showif) {
            re = new RegExp(itemObject.showif, "ig");
            itemObject.si_compiled = re;
        }        
        itemObject.sitarget = cqrParser.xmlGetFirst(node, "showiftarget");        
        
        //Realizing valid <url type=""> tag
        mimes = cqrParser.allowedMIMEs;
        urlObjs = [];
        urlObj = null;
        for (i in mimes) {
            //will use first available valid mime "Url"
            urlObjs = cqrParser.xmlGetElements(node, "Url", mimes[i]);
            if (urlObjs[0]) {
                urlObj = urlObjs[0];
                break;
            }
        }
        
        if (urlObj) {
            //Ok, got one! 
            urlAttrs = cqrParser.xmlGetElementAttrs(urlObj);
            if (urlAttrs) {
                itemObject.Action = urlAttrs["template"] || "";
                itemObject.Method = urlAttrs["method"] || "GET";
            }
            
            //Now looking for "inputs"
            inputs = cqrParser.xmlGetElements(urlObj, 'Param');
            if (inputs.length){
                for (var i=0; i<inputs.length; i++){
                    //... and it's attributes
                    inputAttrs = cqrParser.xmlGetElementAttrs(inputs[i]);
                    if (!inputAttrs) continue;
                    inputAttrs["priority"] = i;
                    //Done! Saving it to query object
                    itemObject._StoreInput(inputAttrs);
                }
            } else {
                ///no inputs
                //if this is non-conquery plugin we'll
                //parse it's template to give to the user
                //a friendly editable inputs.
                if (cqrEnv.ParseNonConQueryUrlTemplate && (plugType != 'ConQuery')){
                    if (itemObject.Action) {
                        var u = cqrParser.urlParser(itemObject.Action);
                        if (u.inputs.length){
                            itemObject.InputsParsedFromUrlTemplate = true;
                            itemObject.Action = u.protocol + '://' + u.domain + u.filepath + u.param;
                            for (i in u.inputs){
                                input = u.inputs[i];
                                input.priority = i;
                                itemObject._StoreInput(input);
                            }
                        }
                    }
                    
                }
            }
        } else {
            ///no valid url tag
        }
        return itemObject;
    },
    
    xmlMozilla: function(node, itemObject, plugType){
        return cqrParser.xmlConQuery(node, itemObject, plugType);
    },
    
    xmlOpenSearch: function(node, itemObject, plugType){
        return cqrParser.xmlConQuery(node, itemObject, plugType);
    },
    
    xmlGetElements: function(doc, strElName, strTypeFilterBy){
        var elts = doc.getElementsByTagName(strElName);
        if (!elts) elts = doc.getElementById(strElname.toLowerCase());
        if (!elts) elts = doc.getElementById(strElname.toUpperCase());
        if (!elts) return '';
        if (!elts[0]) return '';
        if (!strTypeFilterBy) return elts;
        elArr = [];
        for (var i=0; i<elts.length; i++){
            if (elts[i].attributes['type'].value == strTypeFilterBy) elArr.push(elts[i]);
        }
        return elArr;
    },

    xmlGetElementAttrs: function(element){
        //Returns some describing object
        //with all attributes, stored as 
        //atrribute_name:attribute_value
        //or null
        if (!element.attributes.length) return null;
        attrs = [];
        for (var i=0; i<element.attributes.length; i++) {
            attrs[element.attributes[i].name.toLowerCase()] = element.attributes[i].value;
        }
        return attrs;
    },
    
    xmlGetFirst: function(doc, elName, defVal){
        if (!defVal) defVal = '';
        collection = cqrParser.xmlGetElements(doc, elName);
        if (collection && collection[0]) {
            return collection[0].textContent || defVal;
        }else{
            return defVal;
        }
    },
    
    ///////////////////////
    // SRC parsers section
    //
    parseMycroft: function parseMycroft(filedata, itemObject, omit_inputs){
        //var retItem = new cqrProtoQuery();
        if (!filedata) {
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseMycroft got null file data';
            cqrLog.err('cqrParser.parseMycroft got null file data', filepath);
            return itemObject;
        }
        filedata = cqrParser.cleanupMycroft(filedata);
        
        if (itemObject.ext == ".cqr") {
            itemObject.PluginType = 'ConQuery';
        }else{
            itemObject.PluginType = 'Mycroft';
        }
        //SearchTagData = filedata.match(cqrParser.reSearchBody)[0];
        mcSearch = cqrParser.getTags(cqrParser.reSerchTag, filedata);
        if (mcSearch[0]){
            itemObject.Title = mcSearch[0].name || cqrMess.GetStringFromName('cqr.err.parse');
            itemObject.Action = mcSearch[0].action || "http://conquery.mozdev.org/pluginsdocs.html";
            itemObject.Desc = mcSearch[0].description || "";
            itemObject.referrer = mcSearch[0].referrer || "";
            itemObject.opentarget = mcSearch[0].opentarget || "";
            itemObject.openwin = mcSearch[0].openwin || "";
            itemObject.forcein = mcSearch[0].forcein || "";
            //XXX: Leo's malformed plugins special patch:
                p_plgname = itemObject.filepath.split(cqrFS.path_delim);
                if (p_plgname[p_plgname.length-1] == "leo.src") {
                    //because of two "name" params defined:
                    itemObject.Title = "Leo Eng<->Ger";
                    //because of malformed action:
                    itemObject.Action = "http://dict.leo.org/";
                    itemObject.Desc = cqrMess.GetStringFromName('cqr.err.leo') + retItem.Desc;
                } else if (p_plgname[p_plgname.length-1] == "leofrn.src"){
                    //because of two name params defined:
                    itemObject.Title = "Leo Frn<->Amd";
                    //because of malformed action:
                    itemObject.Action = "http://dico.leo.org/";
                    itemObject.Desc = cqrMess.GetStringFromName('cqr.err.leo') + retItem.Desc;
                }
            //EOf leo patch
            itemObject.Method = mcSearch[0].method || "GET";
            itemObject.qcharset = mcSearch[0].querycharset || "UTF-8";
            itemObject.showif = mcSearch[0].showif || ""; 
            if (itemObject.showif) {
                re = new RegExp(itemObject.showif, "ig");
                itemObject.si_compiled = re;
            }
            itemObject.sitarget = mcSearch[0].showiftarget || "[:selection]";
            if (!omit_inputs) {
                var inp = cqrParser.getTags(cqrParser.reInputTag, filedata);
                for (each in inp) itemObject._StoreInput(inp[each]);
            }
            //retItem.eId = cqrParser.composeId(filepath);
            itemObject.ShowInMenu = true;
        }else{
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseMycroft cant find <search ...> tag';
            cqrLog.err('cqrParser.parseMycroft cant find <search ...> tag', filepath);
            return itemObject;
        }
        return itemObject; 
    },
    
    getTags: function(reTag, str){
        //This extracts all <tag param="val< - > ue"> patterns from str
        var TagsData = [];
        var TagsBuff = [];
        var cqrTagsRE = reTag;
        while (TagsBuff!=null){
            TagsBuff = cqrTagsRE.exec(str);
            if (TagsBuff!=null){
                TagsData[TagsData.length]=cqrParser.tagParams(TagsBuff[0]);
                TagsData[TagsData.length-1].priority = cqrTagsRE.lastIndex;
            }
        }
    return TagsData;
    },

    tagParams: function(tag){
      var Params = new Object;
      var ParamsBuff = "  ";
      var ParamsData = [];
      //To correct Mycroft's "user" in <input>.
      tag = String(tag).replace(cqrParser.reUserParam, ' value="[:selection]"');
      Params["value"]= cqrMess.GetStringFromName('cqr.err.parse');
      ParamsBuff = cqrParser.reTagParams.exec(tag);
      while (ParamsBuff!=null){
            if (ParamsBuff!=null){
                ParamsArr = String(ParamsBuff[0]).split(/=\s*\"/);
                ParamsArr[1] = ParamsArr[1].replace(/^\s+/, "");
                Params[String(ParamsArr[0].replace(/\s+/,"")).toLowerCase()] = ParamsArr[1].replace(/\"\s*/, "");
                ParamsBuff = cqrParser.reTagParams.exec(tag);
            }
        }
    return Params;
    },

    cleanupMycroft: function(strData){
        strData = strData.replace(cqrParser.reComments, '');
        strData = strData.replace(cqrParser.reSpecials, '  ');
        return strData;
    },
    
    urlParser: function(strURL) {
        //strURL = 'lingvo.yandex.ru';
        urlObj = new Object;
        urlObj.protocol = '';
        urlObj.domain = '';
        urlObj.path = '';
        
        URLi = Components.classes["@mozilla.org/network/url-parser;1?auth=maybe"].
                       createInstance(Components.interfaces.nsIURLParser);
        var specLen = -1, schemePos = {}, schemeLen = {}, authorityPos = {}, authorityLen = {}, pathPos = {}, pathLen = {};
        URLi.parseURL(strURL, specLen, schemePos, schemeLen, authorityPos, authorityLen, pathPos, pathLen);
        //cqrLog.info(strURL);
        urlObj.protocol = strURL.substr(schemePos.value, schemeLen.value);
        urlObj.domain = strURL.substr(authorityPos.value, authorityLen.value);
        urlObj.path = strURL.substr(pathPos.value, pathLen.value);
        
        var filepathPos = {}, filepathLen = {}, paramPos = {}, paramLen = {}, queryPos = {}, queryLen = {}, refPos = {}, refLen = {};
        URLi.parsePath(urlObj.path, -1, filepathPos, filepathLen, paramPos, paramLen, queryPos, queryLen, refPos, refLen);

        urlObj.filepath = urlObj.path.substr(filepathPos.value, filepathLen.value);
        urlObj.param = urlObj.path.substr(paramPos.value, paramLen.value);
        urlObj.query = urlObj.path.substr(queryPos.value, queryLen.value);
        urlObj.ref = urlObj.path.substr(refPos.value, refLen.value);
        urlObj.inputs = [];
        
        if (!urlObj.query) return urlObj;
        var delim = '';
        if (urlObj.query.indexOf(';')>=0) delim = ';';
        if (urlObj.query.indexOf('&')>=0) delim = '&';
        if (delim){
            var pairs = urlObj.query.split(delim);
            for (i in pairs){
                pair = pairs[i].split('=');
                urlObj.inputs.push({'name':pair[0], 'value':pair[1]});
            }
        } else {
            if (urlObj.query.indexOf('=')>=0) {
                //url has only single pair
                pair = urlObj.query.split('=');
                urlObj.inputs.push({'name':pair[0], 'value':pair[1]});
            } else {
                //single param with no value or malformed url
                urlObj.inputs.push({'name':urlObj.query, 'value':''});
            }
        }
        return urlObj;
    },

}

cqrParser.init();
